# -*- coding: utf-8 -*-
# Copyright (C) 2018  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.

from wendelin.lib.testing import getTestDB
from wendelin.bigfile.file_zodb import ZBigFile
from wendelin.bigfile.tests.test_filezodb import blksize
from wendelin import wcfs

import transaction
from persistent import Persistent
from persistent.timestamp import TimeStamp

import os, os.path, subprocess
from errno import EINVAL
from pytest import raises

testdb = None
testzurl = None
testmntpt = None
def setup_module():
    global testdb, testzurl, testmntpt
    testdb = getTestDB()
    testdb.setup()

    zstor = testdb.getZODBStorage()
    testzurl = wcfs._zstor_2zurl(zstor)
    zstor.close()
    testmntpt = wcfs._mntpt_4zurl(testzurl)
    os.rmdir(testmntpt)

def teardown_module():
    testdb.teardown()

# make sure we start every test without wcfs server running.
def setup_function(f):
    assert not os.path.exists(testmntpt)

# make sure we unmount wcfs after every test.
def teardown_function(f):
    mounted = not subprocess.call(["mountpoint", "-q", testmntpt])
    if mounted:
        subprocess.check_call(["fusermount", "-u", testmntpt])
    if os.path.exists(testmntpt):
        os.rmdir(testmntpt)


# readfile reads file @ path.
def readfile(path):
    with open(path) as f:
        return f.read()

# ttime converts tid to transaction commit time.
def ttime(tid):
    return TimeStamp(tid).timeTime()


# check that zurl does not change from one open to another storage open.
def test_zurlstable():
    for i in range(10):
        zstor = testdb.getZODBStorage()
        zurl  = wcfs._zstor_2zurl(zstor)
        zstor.close()
        assert zurl == testzurl


def test_join():
    zurl = testzurl
    with raises(RuntimeError, match="wcfs: join .*: server not started"):
        wcfs.join(zurl, autostart=False)

    wc = wcfs._start(zurl)
    assert wc.mountpoint == testmntpt
    assert readfile(wc.mountpoint + "/.wcfs") == zurl
    assert os.path.isdir(wc.mountpoint + "/bigfile")

    wc2 = wcfs.join(zurl, autostart=False)
    assert wc2.mountpoint == wc.mountpoint

    wc.close()
    wc2.close()


def test_join_autostart():
    zurl = testzurl
    with raises(RuntimeError, match="wcfs: join .*: server not started"):
        wcfs.join(zurl, autostart=False)

    wc = wcfs.join(zurl, autostart=True)
    assert wc.mountpoint == testmntpt
    assert readfile(wc.mountpoint + "/.wcfs") == zurl
    assert os.path.isdir(wc.mountpoint + "/bigfile")

    wc.close()


# XXX parametrize zblk0, zblk1
# XXX select !wcfs mode so that we prepare data through !wcfs path.
def test_bigfile_empty():
    root = testdb.dbopen()
    root['zfile'] = f = ZBigFile(blksize)

    # NOTE there is no clean way to retrieve tid of just committed transaction
    #      we'll be using last._p_serial as workaround.
    root['last'] = last = Persistent()
    transaction.commit()
    tid1 = last._p_serial

    last._p_changed = 1
    transaction.commit()
    tid2 = last._p_serial

    assert ttime(tid2) > ttime(tid1)

    wc = wcfs.join(testzurl, autostart=True)

    # path to bigfile/ under wcfs
    bigpath = wc.mountpoint + "/bigfile"

    # mkdir to non-BigFile - must be rejected
    with raises(OSError) as exc:
        os.mkdir("%s/%s" % (bigpath, last._p_oid.encode('hex')))
    assert exc.value.errno == EINVAL


    """
    # path to f under wcfs
    fpath = "%s/%s" % (bigpath, f._p_oid.encode('hex'))

    os.mkdir(fpath)
    st = os.stat(fpath + "/head/data")
    assert st.st_size == 0
    assert st.st_mtime == ttime(tid1)

    assert readfile(fpath + "/head/at") == 'txn2'
    # XXX head/at = last txn of whole db
    """


    wc.close()



# TODO pagecache state after loading (via mincore)
