// Copyright (c) 2001, 2002 Zope Foundation and Contributors.
// All Rights Reserved.
//
// Copyright (C) 2018  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This software is subject to the provisions of the Zope Public License,
// Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
// THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
// WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
// FOR A PARTICULAR PURPOSE

package main
// Bits that should be in ZODB		XXX -> zodb

import (
	"context"

	"lab.nexedi.com/kirr/neo/go/zodb"

	pickle "github.com/kisielk/og-rek"
)

// Object is common base for in-process representation of ZODB object.
type Object struct {
	jar	*Connection
	oid	zodb.Oid
	serial	zodb.Tid
}

// PyObject is common base for in-process representation of ZODB Python objects.
type PyObject struct {
	Object

	pyclass pickle.Class // python class of this object
	pystate interface{}  // object state. python passes this to pyclass.__new__().__setstate__()
}

// Connection represents a view of ZODB database.
//
// The view is representing state of ZODB objects as of `at` transaction.
//
// XXX Connection, and {Py}Object methods that relate to it, are not safe for
// modifications from multiple goroutines simultaneously.
//
// XXX ^^^ better must be safe - use case: e.g. prefetch.
type Connection struct {
	stor	zodb.IStorage	// underlying storage
	at	zodb.Tid	// current view of database

	// {} oid -> pyobj
	//
	// rationale:
	//
	// on invalidations: we need to go oid -> pyobj and invalidate it.
	// -> Connection need to keep {} oid -> pyobj.
	// -> we can use that {} when loading a persistent Ref twice to get to the same object.
	//
	// however: if Connection keeps strong link to pyobj, just
	// pyobj.PDeactivate will not fully release pyobj if there are no
	// references to it from other objects:
	//
	//	- deactivate will release pyobj state (ok)
	//	- but there will be still reference from connection `oid -> pyobj` map to this object.
	//
	// -> we can solve it by using "weak" pointers in the map.
	//
	// NOTE we cannot use regular map and arbitrarily manually "gc" entries
	// there periodically: since for a pyobj we don't know whether other
	// objects are referencing it, we can't just remove pyobj's oid from
	// the map - if we do so and there are other live objects that
	// reference pyobj, user code can still reach pyobj via those
	// references. On the other hand, if another, not yet loaded, object
	// also references pyobj and gets loaded, traversing reference from
	// that loaded object will load second copy of pyobj, thus breaking 1
	// object in db <-> 1 live object invariant:
	//
	//	A  →  B  →  C
	//	↓           |
	//      D <--------- - - -> D2 (wrong)
	//
	// - A activate
	// - D activate
	// - B activate
	// - D gc, A still keeps link on D
	// - C activate -> it needs to get to D, but D was removed from objtab
	//   -> new D2 is wrongly created
	//
	// that's why we have to depend on Go's GC to know whether there are
	// still live references left or not. And that in turn means finalizers
	// and thus weak references.
	//
	// some link on the subject:
	// https://groups.google.com/forum/#!topic/golang-nuts/PYWxjT2v6ps
	//
	// NOTE2 finalizers don't run on when they are attached to an object in cycle.
	// Hopefully we don't have cycles with ZBtree/ZBucket	XXX verify this
	objmu  sync.Mutex
	objtab map[zodb.Oid]interface{}	// oid -> WeakRef(PyObject) | loadInProgress
}

// loadInProgress entry in Conn.objtab tells users, that try to get the entry,
// that another goroutine is already in progress of loading it.
type loadInProgress struct {
	ready chan struct{} // closed when loading finishes

	// result of the load
	pyobj interface{} // XXX -> PyObject iface
	err   error
}

// Get returns in-RAM object corresponding to specified ZODB object according to current database view.
//
// If there is already in-RAM object that corresponds to oid, that in-RAM object is returned.
// Otherwise new in-RAM object is created and filled with object's class loaded from the database.
//
// The object's data is not neccessarily loaded after Get returns. Use
// PActivate to make sure the object ifs fully loaded.
func (conn *Connection) Get(ctx context.Context, oid zodb.Oid) (*PyObject, error) {
	// XXX = load raw oid, get its class -> get(pyclass, oid)

	conn.objmu.Lock()		// XXX -> rlock
	objentry := conn.objtab[oid]

	// someone else is already loading the object.
	// we have to wait for that someone to finish the load.
	if load, ok := objentry.(*loadInProgress); ok {
		conn.objmu.Unlock()
		select {
		case <-ctx.Done():
			return nil, ctx.Err()	// XXX err ctx
		case <-load.ready:
			// XXX check pyclass match
			return load.pyobj, err
		}
	}

	// noone else is loading the object. Investigate whether it is simply not
	// in objtab, or it is lost weakref, and if so we have to initiate the load.
	var xobj interface{}
	if objentry != nil {
		wobj := objentry.(*WeakRef)
		xobj = wobj.Get()
	}
	if xpyobj != nil {
		// the object is already there
		conn.objmu.Unlock()
		return xpyobj, nil	// XXX + activate?
	}

	// the object is not there, we have to (re)load it.
	load := &loadInProgress{ready: make(chan struct)}
	conn.objTab[oid] = load
	conn.objmu.Unlock()

	pyobj, err = conn.loadpy(ctx, oid)
	load.pyobj = pyobj
	load.err   = err
	close(load.ready)

	// replace objtab entry with WeakRef
	conn.objmu.Lock()
	defer conn.objmu.Unlock()
	if x := conn.objtab[oid]; x != load {
		panic(fmt.Sprintf("zodb.Conn: objtab[%s]: load finished, but it is: %#v", oid, x)
	}
	if err != nil {
		// it was loading error - remove the entry not to cause uncontrolled objtab growth.
		// if this object is accessed next time - the load will be retried.
		delete(conn.objtab[oid])
	} else {
		conn.objtab[oid] = NewWeakRef(pyobj)
	}
}

// get returns in-RAM object corresponding to specified ZODB object according to current database view.
//
// If there is already in-RAM object that corresponds to oid, that in-RAM object is returned.
// Otherwise new in-RAM object is created and filled with object's class loaded from the database.
//
// The object's data is not neccessarily loaded after get returns. Use
// PActivate to make sure the object ifs fully loaded.
//
// use-case: in ZODB references are (pyclass, oid), so new ghost is created without loading anything.
func (conn *Connection) get(ctx context.Context, pyclass pickle.Class, oid zodb.Oid) (*PyObject, error) {
	conn.objmu.Lock()		// XXX -> rlock
	objentry := conn.objtab[oid]

	// someone else is already loading the object.
	// we have to wait for that someone to finish the load.
	if load, ok := objentry.(*loadInProgress); ok {
		conn.objmu.Unlock()
		select {
		case <-ctx.Done():
			return nil, ctx.Err()	// XXX err ctx
		case <-load.ready:
			// XXX check pyclass match
			return load.pyobj, err
		}
	}

	// noone else is loading the object. Investigate whether it is simply not
	// in objtab, or it is lost weakref, and if so we have to initiate the load.
	var xobj interface{}
	if objentry != nil {
		wobj := objentry.(*WeakRef)
		xobj = wobj.Get()
	}
	if xpyobj != nil {
		// the object is already there
		conn.objmu.Unlock()
		return xpyobj, nil	// XXX + activate?
	}

	// the object is not there, we have to (re)load it.
	load := &loadInProgress{ready: make(chan struct)}
	conn.objTab[oid] = load
	conn.objmu.Unlock()

	pyobj, err = conn.loadpy(ctx, oid)
	load.pyobj = pyobj
	load.err   = err
	close(load.ready)

	// replace objtab entry with WeakRef
	conn.objmu.Lock()
	defer conn.objmu.Unlock()
	if x := conn.objtab[oid]; x != load {
		panic(fmt.Sprintf("zodb.Conn: objtab[%s]: load finished, but it is: %#v", oid, x)
	}
	if err != nil {
		// it was loading error - remove the entry not to cause uncontrolled objtab growth.
		// if this object is accessed next time - the load will be retried.
		delete(conn.objtab[oid])
	} else {
		conn.objtab[oid] = NewWeakRef(pyobj)
	}


/*
	// XXX -> loadpy
	buf, serial, err := conn.stor.Load(ctx, zodb.Xid{Oid: oid, At: conn.at})
	if err != nil {
		return nil, err
	}

	pyclass, pystate, err := zodb.PyData(buf.Data).Decode()
	if err != nil {
		return nil, err	// XXX err ctx
	}

	buf.Release()

	return &PyObject{
		Object:  Object{jar: conn,oid: oid, serial: serial},
		pyclass: pyclass,
		pystate: pystate,
	}, nil
*/
}

func (conn *Connection) load(ctx context.Context, oid zodb.Oid) (*PyObject, error) {
	// XXX
}

func (conn *Connection) loadpy(ctx context.Context, oid zodb.Oid) (pyclass pickle.Class, pystate interface{}, serial zodb.Tid, _ error) {
	buf, serial, err := conn.stor.Load(ctx, zodb.Xid{Oid: oid, At: conn.at})
	if err != nil {
		return pickle.Class{}, nil, 0, err
	}

	pyclass, pystate, err = zodb.PyData(buf.Data).Decode()
	if err != nil {
		return pickle.Class{}, nil, 0, err	// XXX err ctx
	}

	buf.Release()

	return pyclass, pystate, serial, nil
}

// path(class) -> new(pyobj)
var classTab = map[string]func(*PyObject)interface{}

// registerClass registers python class to be transformed to Go instance
// created via classNew.
//
// must be called from global init().
func registerClass(pyClassPath string, classNew func(*PyObject)interface{}) {
	classTab[classPath] = classNew
}

// newGhost creates new ghost object corresponding to pyclass and oid.
func (conn *Connection) newGhost(pyclass pickle.Class, oid zodb.Oid) interface{} {
	pyobj := &PyObject{
		Object:  Object{jar: conn, oid: oid, serial: 0},
		pyclass: pyclass,
		pystate: nil,
	}

	// switch on pyclass and transform e.g. "zodb.BTree.Bucket" -> *ZBucket
	classNew := classTab[pyclass.Module + "." + pyclass.Name]
	if classNew == nil {
		return pyobj	// XXX or return error here?
	}

	return classNew(pyobj)
}


// XXX pyobj.PInvalidate()	= deactivate without checking if state != modified

// PDeactivates transforms object to ghost state.
//
// In ghost state object data is dropped and only oid/pyclass information is left in RAM.
func (pyobj *PyObject) PDeactivate() {
	// FIXME if state=modified PDeactivate must be noop.

	pyobj.pystate = nil
	pyobj.serial  = 0
}

// PActivate brings object to live state.
//
// If object state is not in RAM - it is loaded from the database.
func (pyobj *PyObject) PActivate(ctx context.Context) error {
	if pyobj.pystate != nil {
		return nil // already loaded
	}

	pyclass, pystate, serial, err := pyobj.jar.loadpy(ctx, pyobj.oid)
	if err != nil {
		return err
	}

	if pyclass != pyobj.pyclass {
		// XXX complain pyclass changed
		// XXX both ref and object data uses pyclass so it indeed can be different
	}

	pyobj.serial  = serial
	pyobj.pystate = pystate
	return nil
}
